
#include "config.h"

#define INI_MAX_LINE 4096
#define INI_STOP_ON_FIRST_ERROR 1

#include "ini.h"
#include "def.h"

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <getopt.h>

int m_tty = -1;

// show usage
static void
usage(void)
{
	(void)fprintf(stderr, "usage: dxmenu [-eirw] [-p position] [-c conffile] [title]\n");
	exit(1);
}

/* call strdup checking for error */
static char *
estrdup(const char *s)
{
	char *t;

	if ((t = strdup(s)) == NULL)
		err(1, "strdup");
	return t;
}

/* parse position string from -p, put results on config.posx, config.posy, and config.monitor */
static void
parseposition(struct dxmenu_config* config, char *optarg)
{
	long n;
	char *s = optarg;
	char *endp;

	n = strtol(s, &endp, 10);
	if (errno == ERANGE || n > INT_MAX || n < 0 || endp == s || *endp != 'x')
		goto error;
	config->posx = n;
	s = endp+1;
	n = strtol(s, &endp, 10);
	if (errno == ERANGE || n > INT_MAX || n < 0 || endp == s)
		goto error;
	config->posy = n;
	if (*endp == ':') {
		s = endp+1;
		mflag = 1;
		if (strncasecmp(s, "CUR", 3) == 0) {
			config->monitor = -1;
			endp = s+3;
		} else {
			n = strtol(s, &endp, 10);
			if (errno == ERANGE || n > INT_MAX || n < 0 || endp == s || *endp != '\0')
				goto error;
			config->monitor = n;
		}
	} else if (*endp != '\0') {
		goto error;
	}

	return;

error:
	errx(1, "improper position: %s", optarg);
}

/* set button global variable */
static void
setbutton(char *s)
{
	size_t len;

	if ((len = strlen(s)) < 1)
		return;
	switch (s[len-1]) {
	case '1': button = Button1; break;
	case '2': button = Button2; break;
	case '3': button = Button3; break;
	default:  button = atoi(&s[len-1]); break;
	}
}

/* parse icon path string */
static void
parseiconpaths(char *s)
{
	if (s == NULL)
		return;
	free(iconstring);
	iconstring = estrdup(s);
	niconpaths = 0;
	for (s = strtok(iconstring, ":"); s != NULL; s = strtok(NULL, ":")) {
		if (niconpaths < MAXPATHS) {
			iconpaths[niconpaths++] = s;
		}
	}
}

/* set modifier global variable */
static void
setmodifier(char *s)
{
	size_t len;

	if ((len = strlen(s)) < 1)
		return;
	switch (s[len-1]) {
	case '1': modifier = Mod1Mask; break;
	case '2': modifier = Mod2Mask; break;
	case '3': modifier = Mod3Mask; break;
	case '4': modifier = Mod4Mask; break;
	case '5': modifier = Mod5Mask; break;
	default:
		if (strcasecmp(s, "Alt") == 0) {
			modifier = Mod1Mask;
		} else if (strcasecmp(s, "Super") == 0) {
			modifier = Mod4Mask;
		}
		break;
	}
}

// parse an unsigned 64-bit number 
// uint64_str must contain *only* the text of the number 
// return the number, and set *success to true if we succeeded
// return -EINVAL if any of hte arguments were NULL
// return 0, LLONG_MAX, or LLONG_MIN if we failed
static uint64_t dxmenu_parse_uint64( char const* uint64_str, bool* success ) {
   
   if( uint64_str == NULL || success == NULL ) {
      return -EINVAL;
   }
   
   char* tmp = NULL;
   uint64_t uid = (uint64_t)strtoll( uint64_str, &tmp, 10 );
   
   if( tmp == uint64_str ) {
      *success = false;
   }
   else {
      *success = true;
   }
   
   return uid;
}


// join two paths, writing the result to dest if dest is not NULL.
// otherwise, allocate and return a buffer containing the joined paths.
// return NULL on OOM 
static char* dxmenu_fullpath( char const* root, char const* path, char* dest ) {

   char delim = 0;
   int path_off = 0;
   int len = 0;
   
   if( root == NULL || path == NULL ) {
      return NULL;
   }
   
   len = strlen(path) + strlen(root) + 2;

   if( strlen(root) > 0 ) {
      size_t root_delim_off = strlen(root) - 1;
      if( root[root_delim_off] != '/' && path[0] != '/' ) {
         len++;
         delim = '/';
      }
      else if( root[root_delim_off] == '/' && path[0] == '/' ) {
         path_off = 1;
      }
   }

   if( dest == NULL ) {
      dest = (char*)calloc( sizeof(char) * (len), 1 );
      if( dest == NULL ) {
         return NULL;
      }
   }

   memset(dest, 0, len);

   strcpy( dest, root );
   if( delim != 0 ) {
      dest[strlen(dest)] = '/';
   }
   strcat( dest, path + path_off );

   return dest;
}

// ini parser callback 
// return 1 on parsed (WARNING: masks OOM)
// return 0 on not parsed
static int dxmenu_config_ini_parser( void* userdata, char const* section, char const* name, char const* value ) {
   
   struct dxmenu_config* conf = (struct dxmenu_config*)userdata;
   bool success = false;   
   
   if( strcmp( section, DXMENU_CONFIG_NAME ) == 0 ) {
      
      if( strcmp( name, DXMENU_CONFIG_PATH ) == 0 ) {
         
         if( conf->config_path == NULL ) {
            // save this 
            conf->config_path = strdup_or_null( value );
         }
         return 1;
      }
      
      else if( strcmp( name, DXMENU_CONFIG_JSON_MENU_PATH ) == 0 ) {
         
         if( conf->json_menu_path == NULL ) {
            // save this 
            conf->json_menu_path = strdup_or_null( value );
         }
         return 1;
      }
    
      else if( strcmp( name, DXMENU_CONFIG_FONT ) == 0 ) {
         
         if( conf->font == NULL ) {
            // save this 
            conf->font = strdup_or_null( value );
         }
         return 1;
      }
      
      else if( strcmp( name, DXMENU_CONFIG_BACKGROUND_COLOR ) == 0 ) {
         
         if( conf->background_color == NULL ) {
            // save this 
            conf->background_color = strdup_or_null( value );
         }
         return 1;
      }
      
      else if( strcmp( name, DXMENU_CONFIG_FOREGROUND_COLOR ) == 0 ) {
         
         if( conf->foreground_color == NULL ) {
            // save this 
            conf->foreground_color = strdup_or_null( value );
         }
         return 1;
      }
      
      else if( strcmp( name, DXMENU_CONFIG_SELBACKGROUND_COLOR ) == 0 ) {
         
         if( conf->selbackground_color == NULL ) {
            // save this 
            conf->selbackground_color = strdup_or_null( value );
         }
         return 1;
      }
      
      else if( strcmp( name, DXMENU_CONFIG_SELFOREGROUND_COLOR ) == 0 ) {
         
         if( conf->selforeground_color == NULL ) {
            // save this 
            conf->selforeground_color = strdup_or_null( value );
         }
         return 1;
      }
      
      else if( strcmp( name, DXMENU_CONFIG_SEPARATOR_COLOR ) == 0 ) {
         
         if( conf->separator_color == NULL ) {
            // save this 
            conf->separator_color = strdup_or_null( value );
         }
         return 1;
      }
      
      else if( strcmp( name, DXMENU_CONFIG_BORDER_COLOR ) == 0 ) {
         
         if( conf->border_color == NULL ) {
            // save this 
            conf->border_color = strdup_or_null( value );
         }
         return 1;
      }
            
      else if( strcmp( name, DXMENU_CONFIG_WIDTH_PIXELS ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
        
         long_val = strtoul( value, &tmp, 10 );
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         /*
         else if ((long) someIntMin > long_val || long_val > (long) someIntMax)
            handle_invalid();
         */
         else {
            // save this            
            conf->width_pixels = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_HEIGHT_PIXELS ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->height_pixels = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_BORDER_PIXELS ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->border_pixels = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_SEPARATOR_PIXELS ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->separator_pixels = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_GAP_PIXELS ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->gap_pixels = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_MAX_ITEMS ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->max_items = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_TRIANGLE_WIDTH ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->triangle_width = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_TRIANGLE_HEIGHT ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->triangle_height = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_ICONPADDING ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->iconpadding = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_HORZPADDING ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->horzpadding = (int) long_val;
            return 1;
         }
      }
            
      else if( strcmp( name, DXMENU_CONFIG_ALIGNMENT ) == 0 ) {
      
         char* tmp = NULL;
         long int long_val;
         
         long_val = strtoul( value, &tmp, 10 );
         
         if( *tmp != '\0' ) {
            
            fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
            return 0;
         }
         else {
            // save this            
            conf->alignment = (int) long_val;
            return 1;
         }
      }
      
      else if( strcmp( name, DXMENU_CONFIG_HASICON ) == 0 ) {
         
         if( strcasecmp( name, "true" ) == 0 ) {
            
            conf->hasicon = true;
         }
         else if( strcasecmp( name, "false" ) == 0 ) {
            
            conf->hasicon = false;
         }
         else if( !conf->hasicon ) {
            
            // maybe it's 0 or non-zero?
            conf->hasicon = (bool)dxmenu_parse_uint64( value, &success );
            if( !success ) {
               
               fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
               return 0;
            }
            else {
               
               return 1;
            }
         }
      }
      
      else if( strcmp( name, DXMENU_CONFIG_TINT2EXECP ) == 0 ) {
         
         if( strcasecmp( name, "true" ) == 0 ) {
            
            conf->tint2execp = true;
         }
         else if( strcasecmp( name, "false" ) == 0 ) {
            
            conf->tint2execp = false;
         }
         else if( !conf->tint2execp ) {
            
            // maybe it's 0 or non-zero?
            conf->tint2execp = (bool)dxmenu_parse_uint64( value, &success );
            if( !success ) {
               
               fprintf(stderr, "Invalid value '%s' for '%s'\n", value, name );
               return 0;
            }
            else {
               
               return 1;
            }
         }
      }
      
      else
         fprintf(stderr, "Unrecognized field '%s'\n", name);

      return 1;
   }
   
   return 1;
}

// config sanity check 
int dxmenu_config_sanity_check( struct dxmenu_config* conf ) {
   
   int rc = 0;
   
   if( conf->config_path == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing config_path\n");
      rc = -EINVAL;
   }
   
   if( conf->json_menu_path == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing json_menu_path\n");
      rc = -EINVAL;
   }
   
   if( conf->background_color == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing background_color\n");
      rc = -EINVAL;
   }
   
   if( conf->foreground_color == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing foreground_color\n");
      rc = -EINVAL;
   }
   
   if( conf->selbackground_color == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing selbackground_color\n");
      rc = -EINVAL;
   }
   
   if( conf->selforeground_color == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing selforeground_color\n");
      rc = -EINVAL;
   }
   
   if( conf->border_color == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing border_color\n");
      rc = -EINVAL;
   }
   
   if( conf->separator_color == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing separator_color\n");
      rc = -EINVAL;
   }   
   
   return rc;
}
   

// initialize a config 
// always succeeds
int dxmenu_config_init( struct dxmenu_config* conf )
{
   memset( conf, 0, sizeof(struct dxmenu_config) );
   return 0;
}

// load from a file, by path
// return on on success
// return -errno on failure to open 
int dxmenu_config_load( char const* path, struct dxmenu_config* conf )
{
   
   FILE* f = NULL;
   int rc = 0;
   
   f = fopen(path, "r");
   if(!f) {
      rc = -errno;
      return rc;
   }
   
   rc = dxmenu_config_load_file(f, conf);
   
   fclose(f);
   
   return rc;
}

// load from a file
// return 0 on success
// return -errno on failure to load
int dxmenu_config_load_file( FILE* file, struct dxmenu_config* conf ) {
   
   int rc = 0;
   
   rc = ini_parse_file( file, dxmenu_config_ini_parser, conf );
   if( rc != 0 ) {
      fprintf(stderr, "ini_parse_file(config) rc = %d\n", rc );
      dxmenu_config_free( conf );

      return rc;
   }

   // convert paths 
   rc = dxmenu_config_fullpaths( conf );
   if( rc != 0 ) {

      fprintf(stderr, "dxmenu_config_fullpaths: %s\n", strerror(-rc) );
      dxmenu_config_free( conf );
      return rc;
   }
   
   return rc;
}

// free a config
// always succeeds
int dxmenu_config_free( struct dxmenu_config* conf )
{
   if( conf->config_path != NULL ) {
      
      free( conf->config_path );
      conf->config_path = NULL;
   }
   
   if( conf->json_menu_path != NULL ) {
      
      free( conf->json_menu_path );
      conf->json_menu_path = NULL;
   }
   
   if( conf->font != NULL ) {
      
      free( conf->font );
      conf->font = NULL;
   }
   
   if( conf->background_color != NULL ) {
      
      free( conf->background_color );
      conf->background_color = NULL;
   }
   
   if( conf->foreground_color != NULL ) {
      
      free( conf->foreground_color );
      conf->foreground_color = NULL;
   }
   
   if( conf->selbackground_color != NULL ) {
      
      free( conf->selbackground_color );
      conf->selbackground_color = NULL;
   }
   
   if( conf->selforeground_color != NULL ) {
      
      free( conf->selforeground_color );
      conf->selforeground_color = NULL;
   }
   
   if( conf->separator_color != NULL ) {
      
      free( conf->separator_color );
      conf->separator_color = NULL;
   }
   
   
   
   if( conf->border_color != NULL ) {
      
      free( conf->border_color );
      conf->border_color = NULL;
   }
   
   return 0;
}

// convert all paths in the config to absolute paths 
// return 0 on success 
// return -ENOMEM on OOM 
// return -ERANGE if cwd is too long
int dxmenu_config_fullpaths( struct dxmenu_config* conf ) {
   
   char** need_fullpath[] = {
      &conf->config_path,
      NULL
   };
   
   char cwd_buf[ PATH_MAX + 1 ];
   memset( cwd_buf, 0, PATH_MAX + 1 );
   
   char* tmp = getcwd( cwd_buf, PATH_MAX );
   if( tmp == NULL ) {
      
      fprintf(stderr, "Current working directory exceeds %u bytes\n", PATH_MAX);
      return -ERANGE;
   }
   
   for( int i = 0; need_fullpath[i] != NULL; i++ ) {
      
      if( need_fullpath[i] != NULL && (*need_fullpath[i]) != NULL ) {
         
         if( *(need_fullpath[i])[0] != '/' ) {
               
            // relative path 
            char* new_path = dxmenu_fullpath( cwd_buf, *(need_fullpath)[i], NULL );
            if( new_path == NULL ) {
               
               return -ENOMEM;
            }
            
            free( *(need_fullpath[i]) );
            *(need_fullpath[i]) = new_path;
         }
      }
   }
   
   return 0;
}


/* get configuration from command-line options */
void dxmenu_config_load_from_args(struct dxmenu_config* config, int argc, char *argv[])
{
	int ch;
	char *s, *t;

	classh.res_class = CLASS;
	classh.res_name = argv[0];
	if ((s = strrchr(argv[0], '/')) != NULL)
		classh.res_name = s + 1;
	parseiconpaths(getenv(ICONPATH));
	while ((ch = getopt(argc, argv, "c:eip:rwx:X:o:")) != -1) {
		switch (ch) {
        case 'c':
         {
            if(config->config_path != NULL) {
               free(config->config_path);
            }
            // save this
            config->config_path = strdup_or_null(optarg);
            break;
         }
		case 'e':
			eflag = 1;
			break;
		case 'i':
			iflag = 1;
			break;
		case 'p':
			pflag = 1;
			parseposition(config, optarg);
			break;
		case 'r':
			rflag = 1;
			break;
		case 'w':
			wflag = 1;
			break;
		case 'X':
			passclickflag = 1;
			/* PASSTHROUGH */
		case 'x':
			rootmodeflag = 1;
			s = optarg;
			setbutton(s);
			if ((t = strchr(s, '-')) == NULL)
				return;
			*t = '\0';
			setmodifier(s);
			break;
		case 'o':
			(void)fprintf(stderr, "Redirecting to /dev/pts/%s...\n", optarg);
			m_tty = atoi(optarg);
			break;
		default:
		    usage();
			break;
		}
	}
	if (rootmodeflag)
		wflag = 0;
	argc -= optind;
	argv += optind;
	if (argc != 0)
		usage();	
	return;
}
